#ifndef CONFIGURATION_FILE_H
#define CONFIGURATION_FILE_H

#include <iostream>
#include <chrono>
#include <boost/property_tree/ini_parser.hpp>
#include <kr2_program_api/api_v1/bundles/custom_device.h>

class ConfifurationFile {
	public:
       ConfifurationFile(){
       }

        struct PARSED_DATA{
            // Parameters
            // main
            std::string USERNAME;
            std::string LOG_PATH;
            std::string HOST;
            std::string DIAGNOSTICS_PATH;
            std::string CALIBRATION_PATH;
            // stats
            unsigned int STATS_INTERVAL; //milliseconds
            bool         HALT_TRIGGER ;
            unsigned int MAX_CONSECUTIVE_COMMANDS_LOST;
            unsigned int MAX_RECENT_COMMANDS_LOST;
            unsigned int MAX_RECENT_AVG_SYSTEM_JITTER;   //micros
            unsigned int MAX_RECENT_MAX_SYSTEM_JITTER;   //micros
            unsigned int MAX_RECENT_AVG_ROUNDTRIP_TIME; //micros
            unsigned int MAX_RECENT_AVG_CMD_JITTER;         //micros
            double JOINT_MAX_DIFF_WHEN_MOVE_ENDED;
            double LINEAR_MAX_DIFF_WHEN_MOVE_ENDED;
            // debug
            bool CPU_READER_ENABLE;
        };

        PARSED_DATA get_data(){
            return parsed_data_;
        }

       void chooseFile(std::string a_config_file_path, std::string a_config_file_name = "KORD.ini"){
            config_file_path_ = a_config_file_path;
            config_file_name_ = a_config_file_name;
       }

       void read(){
            std::string config = config_file_path_ + config_file_name_;

            boost::property_tree::ptree pt;
            try { boost::property_tree::ini_parser::read_ini(config, pt); }
            catch (std::exception &e)
            {
                // problem with reading the KORD.ini
                CLOG_ERR("Problem with reading KORD.ini");

                // setting defaults
                //main
                parsed_data_.USERNAME = DEF_USERNAME;
                parsed_data_.LOG_PATH = DEF_LOG_PATH;
                parsed_data_.HOST = DEF_HOST;
                parsed_data_.DIAGNOSTICS_PATH = DEF_DIAGNOSTICS_PATH;
                parsed_data_.CALIBRATION_PATH = DEF_CALIBRATION_PATH;
                //stats
                parsed_data_.STATS_INTERVAL = DEF_STATS_INTERVAL_; //milliseconds
                parsed_data_.HALT_TRIGGER    = DEF_HALT_TRIGGER;
                parsed_data_.MAX_CONSECUTIVE_COMMANDS_LOST = DEF_MAX_CONSECUTIVE_COMMANDS_LOST;
                parsed_data_.MAX_RECENT_COMMANDS_LOST = DEF_MAX_RECENT_COMMANDS_LOST;
                parsed_data_.MAX_RECENT_AVG_SYSTEM_JITTER = DEF_MAX_RECENT_AVG_SYSTEM_JITTER;
                parsed_data_.MAX_RECENT_MAX_SYSTEM_JITTER = DEF_MAX_RECENT_MAX_SYSTEM_JITTER;
                parsed_data_.MAX_RECENT_AVG_ROUNDTRIP_TIME = DEF_MAX_RECENT_AVG_ROUNDTRIP_TIME;
                parsed_data_.MAX_RECENT_AVG_CMD_JITTER = DEF_MAX_RECENT_AVG_CMD_JITTER;
                parsed_data_.JOINT_MAX_DIFF_WHEN_MOVE_ENDED = DEF_JOINT_MAX_DIFF_WHEN_MOVE_ENDED;
                parsed_data_.LINEAR_MAX_DIFF_WHEN_MOVE_ENDED = DEF_LINEAR_MAX_DIFF_WHEN_MOVE_ENDED;

                parsed_data_.CPU_READER_ENABLE = DEF_CPU_READER_ENABLE;

                return;
                // setting defaults and exit
            }

            readFieldFromFile(pt, parsed_data_.USERNAME, DEF_USERNAME, "KORD.username");
            readFieldFromFile(pt, parsed_data_.LOG_PATH, DEF_LOG_PATH, "KORD.log_path");
            readFieldFromFile(pt, parsed_data_.HOST, DEF_HOST, "KORD.host");
            readFieldFromFile(pt, parsed_data_.DIAGNOSTICS_PATH, DEF_DIAGNOSTICS_PATH, "KORD.diagnostics_path");
            readFieldFromFile(pt, parsed_data_.CALIBRATION_PATH, DEF_CALIBRATION_PATH, "KORD.calibration_path");

            readStatWindowFieldFromFile(pt, parsed_data_.STATS_INTERVAL, DEF_STATS_INTERVAL_, "QOC_MONITOR.stat_window_size");
            readEnabledFieldFromFile(pt, parsed_data_.HALT_TRIGGER, DEF_HALT_TRIGGER, "QOC_HALT_TRIGGER.enabled");
            readFieldFromFile(pt, parsed_data_.MAX_CONSECUTIVE_COMMANDS_LOST, DEF_MAX_CONSECUTIVE_COMMANDS_LOST, "QOC_HALT_TRIGGER.on_consecutive_commands_lost");
            readFieldFromFile(pt, parsed_data_.MAX_RECENT_COMMANDS_LOST, DEF_MAX_RECENT_COMMANDS_LOST, "QOC_HALT_TRIGGER.on_recent_commands_lost");
            readFieldFromFile(pt, parsed_data_.MAX_RECENT_AVG_SYSTEM_JITTER, DEF_MAX_RECENT_AVG_SYSTEM_JITTER, "QOC_HALT_TRIGGER.on_recent_avg_system_jitter");
            readFieldFromFile(pt, parsed_data_.MAX_RECENT_MAX_SYSTEM_JITTER, DEF_MAX_RECENT_MAX_SYSTEM_JITTER, "QOC_HALT_TRIGGER.on_recent_max_system_jitter");
            readFieldFromFile(pt, parsed_data_.MAX_RECENT_AVG_ROUNDTRIP_TIME, DEF_MAX_RECENT_AVG_ROUNDTRIP_TIME, "QOC_HALT_TRIGGER.on_recent_avg_roundtrip_time");
            readFieldFromFile(pt, parsed_data_.MAX_RECENT_AVG_CMD_JITTER, DEF_MAX_RECENT_AVG_CMD_JITTER, "QOC_HALT_TRIGGER.on_recent_avg_cmd_jitter");
            readFieldFromFile(pt, parsed_data_.JOINT_MAX_DIFF_WHEN_MOVE_ENDED, DEF_JOINT_MAX_DIFF_WHEN_MOVE_ENDED, "QOC_HALT_TRIGGER.on_movej_cmd_ceased_span");
            readFieldFromFile(pt, parsed_data_.LINEAR_MAX_DIFF_WHEN_MOVE_ENDED, DEF_LINEAR_MAX_DIFF_WHEN_MOVE_ENDED, "QOC_HALT_TRIGGER.on_movel_cmd_ceased_span");

            readEnabledFieldFromFile(pt, parsed_data_.CPU_READER_ENABLE, DEF_CPU_READER_ENABLE, "DEBUG.cpu_reader_enable");
       }


    private:
        std::string config_file_path_;
        std::string config_file_name_;

        //Setup
        //
        const unsigned int MILLISECS_IN_SECOND_ = 1000;

        // Default Parameters
        // main
        const std::string DEF_USERNAME = "";
        const std::string DEF_LOG_PATH = "~/kassow_logs/";
        const std::string DEF_HOST = "";
        const std::string DEF_DIAGNOSTICS_PATH = "~/kassow_diagnostics/";
        const std::string DEF_CALIBRATION_PATH = "~/kassow_calibration/";
        // stats
        const unsigned int DEF_STATS_INTERVAL_ = 200; //milliseconds
        const bool         DEF_HALT_TRIGGER    = true;
        const unsigned int DEF_MAX_CONSECUTIVE_COMMANDS_LOST = 2;
        const unsigned int DEF_MAX_RECENT_COMMANDS_LOST = 5;
        const unsigned int DEF_MAX_RECENT_AVG_SYSTEM_JITTER = 300;
        const unsigned int DEF_MAX_RECENT_MAX_SYSTEM_JITTER = 500;
        const unsigned int DEF_MAX_RECENT_AVG_ROUNDTRIP_TIME = 500;
        const unsigned int DEF_MAX_RECENT_AVG_CMD_JITTER = 500;
        const double DEF_JOINT_MAX_DIFF_WHEN_MOVE_ENDED = 1e-6;
        const double DEF_LINEAR_MAX_DIFF_WHEN_MOVE_ENDED = 1e-6;
        // debug
        const bool DEF_CPU_READER_ENABLE    = false;


        PARSED_DATA parsed_data_;

    // passed as double in seconds, need to convert to milliseconds and integer
    void readStatWindowFieldFromFile(const boost::property_tree::ptree& pt, unsigned int& a_read_value, const unsigned int& a_default_value, std::string a_field_path){
        try { 
            a_read_value = pt.get<double>(a_field_path) * double(MILLISECS_IN_SECOND_);
            CLOG_INFO("Read value: " << a_field_path << " " << std::to_string(a_read_value));
        }
        catch (std::exception &e)
        {
            // can't read the field
            CLOG_ERR("Can't read or find " << a_field_path << " field");
            a_read_value = a_default_value;
        }
    }

    // passed as yes or no, need to convert to bool
    void readEnabledFieldFromFile(const boost::property_tree::ptree& pt, bool& a_read_value, const bool& a_default_value, std::string a_field_path){
        try {
            std::string read_coded_ = pt.get<std::string>(a_field_path);
            if (read_coded_ == "yes"){
                a_read_value = true;
            }
            else{
                if (read_coded_ == "no"){
                    a_read_value = false;
                }
                else{ // not known
                   throw std::runtime_error("Not known value");
                }
            }
            CLOG_INFO("Read value: " << a_field_path << " " << std::to_string(a_read_value));
        }
        catch (std::exception &e)
        {
            // can't read the field
            CLOG_ERR("Can't read or find " << a_field_path << " field");
            a_read_value = a_default_value;
        }
    }

    void readFieldFromFile(const boost::property_tree::ptree& pt, unsigned int& a_read_value, const unsigned int& a_default_value, std::string a_field_path){
        try { 
            a_read_value = pt.get<int>(a_field_path);
            CLOG_INFO("Read value: " << a_field_path << " " << std::to_string(a_read_value));
        }
        catch (std::exception &e)
        {
            // can't read the field
            CLOG_ERR("Can't read or find " << a_field_path << " field");
            a_read_value = a_default_value;
        }
    }

    void readFieldFromFile(const boost::property_tree::ptree& pt, double& a_read_value, const double& a_default_value, std::string a_field_path){
        try { 
            a_read_value = pt.get<double>(a_field_path);
            CLOG_INFO("Read value: " << a_field_path << " " << std::to_string(a_read_value));
        }
        catch (std::exception &e)
        {
            // can't read the field
            CLOG_ERR("Can't read or find " << a_field_path << " field");
            a_read_value = a_default_value;
        }
    }

    void readFieldFromFile(const boost::property_tree::ptree& pt, std::string& a_read_value, const std::string& a_default_value, std::string a_field_path){
        try { 
            a_read_value = pt.get<std::string>(a_field_path);
            CLOG_INFO("Read value: " << a_field_path << " " << a_read_value);
        }
        catch (std::exception &e)
        {
            // can't read the field
            CLOG_ERR("Can't read or find " << a_field_path << " field");
            a_read_value = a_default_value;
        }
    }

};

#endif